-- cpu.vhd: Simple 16bit CPU
-- Copyright (C) 2007 Brno University of Technology,
--                    Faculty of Information Technology
-- Author(s): Zdenek Vasicek <vasicek AT fit.vutbr.cz>
-- vim: set ts=3 et tw=80:
-- id:  999CF0F600A1
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

-- ----------------------------------------------------------------------------
--                        Entity declaration
-- ----------------------------------------------------------------------------
entity cpu is
 port (
   RESET : in std_logic;
   CLK   : in std_logic;
   CE    : in std_logic;
 
   -- PORT
   DADDR : out std_logic_vector(7 downto 0);
   DOUT  : out std_logic_vector(15 downto 0);
   DWE   : out std_logic;

   -- RAM bus
   ABUS  : out std_logic_vector(15 downto 0);   -- address bus
   DBUS  : inout std_logic_vector(15 downto 0); -- data bus
   EN    : out std_logic; -- enable
   RDWR  : out std_logic  -- read / write 

 );
end cpu;


-- ----------------------------------------------------------------------------
--                      Architecture declaration
-- ----------------------------------------------------------------------------
architecture behavioral of cpu is

type fsm_state is (sidle, sfetch0, sfetch1, sdecode, sbranch, shalt, snop, 
                   smload, snegate, smadd0, smadd1, sdload0, sdload1, sdstore, 
                   siload0, siload1, siload2, siload3, sistore0, sistore1, 
                   sistore2, saccdec, saccinc, sijump0, sijump1, 
                   sown, sdataout);

type inst_type is (halt, negate, mload, dload, iload, dstore, istore, nop, 
                   branch, brzero, brpos, brneg, ijump, madd, accdec, accinc,
                   iown, dataout);

type aluoper_type is (alu_add, alu_movb, alu_own);

signal pstate    : fsm_state;
signal nstate    : fsm_state;

signal ireg_reg  : std_logic_vector(15 downto 0);
signal ireg_ld   : std_logic;
signal ireg_dec  : inst_type;
signal ireg_sgnext : std_logic_vector(3 downto 0);
signal ireg_abus : std_logic;

signal pc_reg    : std_logic_vector(15 downto 0);
signal pc_ld     : std_logic;
signal pc_inc    : std_logic;
signal pc_mx     : std_logic_vector(15 downto 0);
signal pc_mx_sel : std_logic;
signal pc_abus   : std_logic;

signal iar_reg   : std_logic_vector(15 downto 0);
signal iar_ld    : std_logic;
signal iar_abus  : std_logic;

signal acc_reg   : std_logic_vector(15 downto 0);
signal acc_ld    : std_logic;
signal acc_mx    : std_logic_vector(15 downto 0);
signal accmx_sel : std_logic_vector(1 downto 0);
signal acc_zero  : std_logic;
signal acc_neg   : std_logic;
signal acc_pos   : std_logic;

signal alu_out     : std_logic_vector(15 downto 0);
signal alu_op0     : std_logic_vector(15 downto 0);
signal alu_mx1_sel : std_logic_vector(1 downto 0);
signal alu_op1     : std_logic_vector(15 downto 0);
signal alu_mx2_sel : std_logic;
signal alu_oper    : aluoper_type;

signal dbus_sel   : std_logic;

begin

   --PORT
   DOUT  <= 
   --== nahradte kod na nasledujicim radku (vyuzijte signal acc_reg) ==
         acc_reg;
   --== na predchozi radek uvedte vas kod ==

   DADDR <= 
   --== nahradte kod na nasledujicim radku (vyuzijte signal ireg_reg) ==
         ireg_reg(7 downto 0);
   --== na predchozi radek uvedte vas kod ==

   --DBUS buffer
   DBUS <= acc_reg when dbus_sel='1' else (others=>'Z');

   --Program counter (PC)
   pc_cntr: process(RESET, CLK)
   begin
      if (RESET='1') then
         pc_reg <= (others=>'0');
      elsif (CLK'event) and (CLK='1') then
         if (pc_ld='1') then
            pc_reg <= pc_mx;
         elsif (pc_inc='1') then
            pc_reg <= pc_reg + 1;
         end if;
      end if;
   end process;

   --PC ABUS buffer
   ABUS <= pc_reg when pc_abus='1' else
           (others => 'Z');

   --PC input multiplexor
   pc_mx <= "0000" & ireg_reg(11 downto 0) when pc_mx_sel='0' else
            DBUS when pc_mx_sel='1';

   --Instruction register IREG
   ireg: process(RESET, CLK)
   begin
      if (RESET='1') then
         ireg_reg <= (others=>'0');
      elsif (CLK'event) and (CLK='1') then
         if (ireg_ld='1') then
            ireg_reg <= DBUS;
         end if;
      end if;
   end process;

   --IREG ABUS buffer
   ABUS <= "0000" & ireg_reg(11 downto 0) when ireg_abus='1' else
           (others => 'Z');

   --Instruction decoder
   process (ireg_reg)
   begin
      case (ireg_reg(15 downto 12)) is 
         when X"0" =>
            case (ireg_reg(3 downto 0)) is
               when X"0"   => ireg_dec <= halt;
               when X"1"   => ireg_dec <= negate;
               when X"2"   => ireg_dec <= accdec;
               when X"3"   => ireg_dec <= accinc;
               --== na nasledujici radek doplnte vas kod ==
               when X"9"   => ireg_dec <= iown;
               --== pred tuto znacku uvedte vas kod ==
               when X"F"   => ireg_dec <= nop;
               when others => ireg_dec <= halt;
            end case;
         when X"1" => ireg_dec <= mload;
         when X"2" => ireg_dec <= dload;
         when X"3" => ireg_dec <= iload;
         when X"4" => ireg_dec <= dstore;
         when X"5" => ireg_dec <= istore;
         when X"6" => ireg_dec <= branch;
         when X"7" => ireg_dec <= brzero;
         when X"8" => ireg_dec <= brpos;
         when X"9" => ireg_dec <= brneg;
         when X"A" => ireg_dec <= madd;
         --== na nasledujici radek doplnte vas kod ==
         when X"B" => ireg_dec <= dataout;
         --== pred tuto znacku uvedte vas kod ==
         when X"F" => ireg_dec <= ijump;
         when others => ireg_dec <= halt;
      end case;
   end process;

   --Indirect address register IAR
   iar: process(RESET, CLK)
   begin
      if (RESET='1') then
         iar_reg <= (others=>'0');
      elsif (CLK'event) and (CLK='1') then
         if (iar_ld='1') then
            iar_reg <= DBUS;
         end if;
      end if;
   end process;

   --IAR ABUS buffer
   ABUS <= iar_reg when iar_abus='1' else
           (others => 'Z');

   --Accumulator register ACC
   accreg: process(RESET, CLK)
   begin
      if (RESET='1') then
         acc_reg <= (others=>'0');
      elsif (CLK'event) and (CLK='1') then
         if (acc_ld='1') then
            acc_reg <= acc_mx;
         end if;
      end if;
   end process;

   --ACC data multiplexor
   acc_mx <= ireg_sgnext & ireg_reg(11 downto 0) when accmx_sel="00" else
             DBUS when accmx_sel="01" else
             alu_out;

   --IREG sign extension
   ireg_sgnext <= "1111" when ireg_reg(11)='1' else 
                  "0000";

   --Arithmetic-logic unit ALU
   --Operands MXs
   alu_op0 <= DBUS when alu_mx1_sel="00" else
              "0000000000000001" when alu_mx1_sel="01" else
              "1111111111111111";
   alu_op1 <= acc_reg when alu_mx2_sel='0' else 
              (not acc_reg);

   --ALU
   with alu_oper select
      alu_out <= alu_op0 + alu_op1 when alu_add,
                 --== na nasledujici radek doplnte vas kod ==
                 alu_op1 - X"001D" when alu_own,
                 --== pred tuto znacku uvedte vas kod ==
                 alu_op1 when others;

   --ACC flags comparator
   acc_zero <= '1' when (acc_reg = "0000000000000000") else '0';
   acc_neg  <= '1' when (acc_reg(15) = '1') else '0';
   acc_pos  <= '1' when (acc_reg(15) = '0') and (acc_zero='0') else '0';

   --FSM present state
   fsm_pstate: process(RESET, CLK)
   begin
      if (RESET='1') then
         pstate <= sidle;
      elsif (CLK'event) and (CLK='1') then
         if (CE = '1') then
            pstate <= nstate;
         end if;
      end if;
   end process;

   --FSM next state logic, Output logic (Moore FSM)
   nsl: process(pstate, ireg_dec, acc_zero, acc_pos, acc_neg)
   begin
      nstate <= sidle;

      EN <= '0';
      DWE <= '0';
      RDWR <= '1';
      ireg_ld <= '0';
      ireg_abus <= '0';
      pc_inc <= '0';
      pc_ld  <= '0';
      pc_abus <= '0';
      pc_mx_sel <= '0';
      iar_ld <= '0';
      iar_abus <= '0';
      accmx_sel <= "00";
      acc_ld <= '0';
      alu_mx1_sel <= "01";
      alu_mx2_sel <= '0';
      alu_oper <= alu_add;
      dbus_sel <= '0';

      case pstate is
         -- IDLE
         when sidle =>
            nstate <= sfetch0;

         -- INSTRUCTION FETCH
         when sfetch0 =>
            nstate <= sfetch1;
            pc_abus <= '1';
            EN <= '1';

         when sfetch1 =>
            ireg_ld <= '1';
            nstate <= sdecode;

         -- INSTRUCTION DECODE
         when sdecode =>
            case ireg_dec is
               when halt =>
                  nstate <= shalt;

               when branch =>
                  nstate <= sbranch;

               when nop =>
                  nstate <= snop;

               when brzero =>
                  if (acc_zero='1') then
                     nstate <= sbranch;
                  else 
                     nstate <= snop;
                  end if;

               when brpos =>
                  if (acc_pos='1') then
                     nstate <= sbranch;
                  else 
                     nstate <= snop;
                  end if;

               when brneg =>
                  if (acc_neg='1') then
                     nstate <= sbranch;
                  else 
                     nstate <= snop;
                  end if;

               when mload =>
                  nstate <= smload;

               when accdec =>
                  nstate <= saccdec;

               when accinc =>
                  nstate <= saccinc;

               when negate =>
                  nstate <= snegate;
               
               when madd =>
                  nstate <= smadd0;

               when dload =>
                  nstate <= sdload0;

               when dstore =>
                  nstate <= sdstore;

               when iload =>
                  nstate <= siload0;

               when istore =>
                  nstate <= sistore0;

               when ijump =>
                  nstate <= sijump0;

               --== od nasledujiciho radku doplnte vas kod ==
               when iown =>
                  nstate <= sown;

               when dataout =>
                  nstate <= sdataout;
               --== pred tuto znacku uvedte vas kod ==

               when others =>
                  nstate <= shalt;

            end case;
         
         -- HALT
         when shalt =>
            nstate <= shalt;

         -- BRANCH
         when sbranch =>
            nstate <= sfetch0;
            pc_ld <= '1';

         -- NOP
         when snop =>
            nstate <= sfetch0;
            pc_inc <= '1';

         -- LOAD IMMEDIATE
         when smload =>
            nstate <= sfetch0;
            accmx_sel <= "00";
            acc_ld <= '1';
            pc_inc <= '1';
            
         -- NEGATE
         when snegate =>
            nstate <= sfetch0;
            accmx_sel <= "10";
            alu_oper <= alu_add;
            alu_mx1_sel <= "01";
            alu_mx2_sel <= '1';
            acc_ld <= '1';
            pc_inc <= '1';

         -- ACC DEC
         when saccdec =>
            nstate <= sfetch0;
            accmx_sel <= "10";
            alu_oper <= alu_add;
            alu_mx1_sel <= "11";
            alu_mx2_sel <= '0';
            acc_ld <= '1';
            pc_inc <= '1';

         -- ACC INC
         when saccinc =>
            nstate <= sfetch0;
            accmx_sel <= "10";
            alu_oper <= alu_add;
            alu_mx1_sel <= "01";
            alu_mx2_sel <= '0';
            acc_ld <= '1';
            pc_inc <= '1';

         -- ADD
         when smadd0 => --phase 0
            nstate <= smadd1;
            ireg_abus <= '1';
            EN <= '1';

         when smadd1 => --phase 1
            nstate <= sfetch0;
            alu_oper <= alu_add;
            accmx_sel <= "10";
            alu_mx1_sel <= "00";
            alu_mx2_sel <= '0';
            acc_ld <= '1';
            pc_inc <= '1';

         -- LOAD DIRECT
         when sdload0 => --phase 0
            nstate <= sdload1;
            ireg_abus <= '1';
            EN <= '1';

         when sdload1 => --phase 1
            nstate <= sfetch0;
            accmx_sel <= "01";
            acc_ld <= '1';
            pc_inc <= '1';

         -- STORE DIRECT
         when sdstore =>
            nstate <= sfetch0;
            dbus_sel <= '1';
            ireg_abus <= '1';
            EN <= '1';
            RDWR <= '0';
            pc_inc <= '1';

         -- LOAD INDIRECT
         when siload0 => --phase 0
            nstate <= siload1;
            ireg_abus <= '1';
            EN <= '1';

         when siload1 => --phase 1
            nstate <= siload2;
            iar_ld <= '1';

         when siload2 => --phase 2
            nstate <= siload3;
            iar_abus <= '1';
            EN <= '1';

         when siload3 => --phase 3
            nstate <= sfetch0;
            accmx_sel <= "01";
            acc_ld <= '1';
            pc_inc <= '1';

         -- STORE INDIRECT
         when sistore0 => --phase 0
            nstate <= sistore1;
            ireg_abus <= '1';
            EN <= '1';

         when sistore1 => --phase 1
            nstate <= sistore2;
            iar_ld <= '1';

         when sistore2 => --phase 2
            nstate <= sfetch0;
            iar_abus <= '1';
            dbus_sel <= '1';
            EN <= '1';
            RDWR <= '0';
            pc_inc <= '1';

         --== od nasledujiciho radku doplnte vas kod ==
         when sown =>
            nstate <= sfetch0;
            accmx_sel <= "10";
            alu_oper <= alu_own;
            alu_mx1_sel <= "11";
            alu_mx2_sel <= '0';
            acc_ld <= '1';
            pc_inc <= '1';

         when sdataout =>
            nstate <= sfetch0;
            DWE <= '1';
            pc_inc <= '1';
         --== pred tuto znacku uvedte vas kod ==

         -- INDIRECT JUMP
         when sijump0 => --phase 0
            nstate <= sijump1;
            ireg_abus <= '1';
            EN <= '1';

         when sijump1 => --phase 1
            nstate <= sfetch0;
            iar_ld <= '1';
            pc_ld <= '1';
            pc_mx_sel <= '1';
            
         when others =>
      end case;
   end process;

end behavioral;

